from collections import OrderedDict

from .exceptions import OutOfContextNodeError
from .types import Source
from .utils import get_coords_of_str_index, get_line


class YamlNode:
    def __init__(self, pnode, level=None, parent=None, comments=()):
        self.pnode = pnode
        self._level = None
        self.parent = parent
        self.comments = list(comments)
        self.level = level

    @property
    def level(self):
        return self._level

    @level.setter
    def level(self, level):
        self._set_level(level)

    def _set_level(self, level):
        self._level = level

    def as_data(self, filename='', raw=False):
        raise NotImplementedError()

    def get_tip(self):
        return self

    def can_add_node(self, node):
        return False

    def add_node(self, node):
        raise NotImplementedError()

    def incorporate_node(self, node):
        if self.can_add_node(node):
            return self.add_node(node)
        else:
            if self.parent is not None:
                return self.parent.incorporate_node(node)
            else:
                # Shouldn't ever get here:
                self.fail_to_incorporate_node(node)  # pragma: no cover

    def fail_to_incorporate_node(self, node):
        pnode = node.pnode
        pos = get_coords_of_str_index(pnode.full_text, pnode.start)
        line = get_line(pnode.full_text, pos.line)
        raise OutOfContextNodeError("Line %s, column %s:\n%s" % (pos.line, pos.column, line))


class ContainerNode(YamlNode):
    def __init__(self, pnode, value=None, **kwargs):
        self.value = value
        super().__init__(pnode, **kwargs)

    def _set_level(self, level):
        self._level = level
        if isinstance(self.value, YamlNode):
            self.value.level = level

    def as_data(self, filename='', raw=False):
        if isinstance(self.value, YamlNode):
            return self.value.as_data(filename, raw=raw)
        else:
            # Shouldn't ever get here.
            return self.value  # pragma: no cover

    def get_tip(self):
        if self.value is not None and isinstance(self.value, YamlNode):
            return self.value.get_tip()
        else:
            return self

    def incorporate_node(self, node):
        if self.can_add_node(node):
            intermediary = None
            if isinstance(node, KeyValue):
                intermediary = Mapping(node.pnode)
            elif isinstance(node, ListItem):
                intermediary = List(node.pnode)

            if intermediary is not None:
                intermediary.level = node.level
                intermediary = self.incorporate_node(intermediary)
                return intermediary.incorporate_node(node)
            else:
                return super().incorporate_node(node)
        else:
            if self.parent is not None:
                return self.parent.incorporate_node(node)
            else:
                self.fail_to_incorporate_node(node)

    def can_add_node(self, node):
        return (
            self.value is None
            and (
                node.level is None
                or node.level > self.level
            )
        )

    def add_node(self, node):
        self.value = node
        node.parent = self
        if node.level is None:
            node.level = self.level
        return node.get_tip()


class ParentNode(YamlNode):
    def __init__(self, pnode, children=(), **kwargs):
        self.children = list(children)
        super().__init__(pnode, **kwargs)

    def _set_level(self, level):
        self._level = level
        for child in self.children:
            if isinstance(child, YamlNode):
                child.level = level

    def get_tip(self):
        if self.children and isinstance(self.children[-1], YamlNode):
            return self.children[-1].get_tip()
        else:
            return self

    def can_add_node(self, node):
        return node.level is None or node.level >= self.level

    def add_node(self, node):
        self.children.append(node)
        node.parent = self
        if node.level is None:
            node.level = self.level
        return node.get_tip()


class Root(ContainerNode):
    def __init__(self, pnode):
        super().__init__(pnode, level=0)

    def can_add_node(self, node):
        return (
            self.value is None
            and (
                node.level is None
                or node.level >= self.level
            )
        )


class Comment(ContainerNode):
    pass


class List(ParentNode):
    def can_add_node(self, node):
        return (
            super().can_add_node(node)
            and
            isinstance(node, ListItem)
        )

    def as_data(self, filename='', raw=False):
        return [c.as_data(filename, raw=raw) for c in self.children]


class ListItem(ContainerNode):
    pass


class Mapping(ParentNode):
    def can_add_node(self, node):
        return (
            super().can_add_node(node)
            and
            isinstance(node, KeyValue)
        )

    def as_data(self, filename='', raw=False):
        return OrderedDict([
            (c.key.as_data(filename, raw=raw), c.as_data(filename, raw=raw))
            for c in self.children
        ])


class KeyValue(ContainerNode):
    def __init__(self, pnode, key, **kwargs):
        self.key = key
        super().__init__(pnode, **kwargs)


class LeafNode(YamlNode):
    def __init__(self, pnode, source_text, value=None, **kwargs):
        self.source_text = source_text
        self.value = [(pnode, value)] if value is not None else [(pnode, source_text)]
        super().__init__(pnode, **kwargs)

    def get_value(self):
        return '\n'.join([v[1] for v in self.value])

    def as_data(self, filename='', raw=False):
        if raw:
            return self.get_value()
        else:
            start_pnode = self.value[0][0]
            end_pnode = self.value[-1][0]
            start = get_coords_of_str_index(start_pnode.full_text, start_pnode.start)
            end = get_coords_of_str_index(end_pnode.full_text, end_pnode.end)

            return Source(
                filename = filename,
                start = start,
                end = end,
                text = self.source_text,
                # Values correspond to lines of text
                value = self.get_value(),
            )

    def can_add_node(self, node):
        return (
            isinstance(node, LeafNode)
            and (
                node.level is None
                or node.level >= self.level
            )
        )

    def add_node(self, node):
        self.value.extend(node.value)
        node.parent = self
        return self.get_tip()


class RawValueLeafNode(LeafNode):
    def get_value(self):
        return self.value[0][1]



import re

from .types import Pos, Source


def get_coords_of_str_index(s, index):
    """Get (line_number, col) of `index` in `string`.

    Based on http://stackoverflow.com/a/24495900
    """
    lines = s.splitlines(True)
    curr_pos = 0
    for linenum, line in enumerate(lines):
        if curr_pos + len(line) > index:
            return Pos(index, linenum + 1, index - curr_pos)
        curr_pos += len(line)
    return Pos(len(s), linenum + 1, 0)


def get_line(s, line_number):
    try:
        return s.splitlines(True)[line_number - 1]
    except IndexError:
        return ''


def get_text_source(text, substring, source_text=None, value=None, filename=''):
    match = re.search(substring, text)
    source_text = match.group() if source_text is None else source_text
    return Source(
        filename = filename,
        start = get_coords_of_str_index(text, match.start()),
        end = get_coords_of_str_index(text, match.end()),
        text = source_text,
        value = value if value is not None else source_text,
    )



import textwrap


syml_base_grammar = textwrap.dedent("""
    lines       = line*
    line        = indent (comment / blank / structure / value) &eol
    structure   = list_item / key_value / section

    indent      = ~"\\s*"

    blank       = &eol
    comment     = ~"(#|//+)+" text?

    list_item   = "-" ws value

    key_value   = section ws data
    section     = key ":"
    key         = ~"[^\\s:]+"

    eol         = "\\n" / ~"$"
    ws          = ~"[ \\t]+"
    text        = ~".+"

    value       = structure / data

""")


text_only_syml_grammar = syml_base_grammar + textwrap.dedent("""
    data        = text
""")

# If we want mare YAML-like booleans:
boolean_syml_grammar = syml_base_grammar + textwrap.dedent("""
    truthy      = "yes" / "Yes" / "YES" / "y" / "Y" / "true" / "True" / "TRUE" / "on" / "On" / "ON"
    falsey      = "no" / "No" / "NO" / "N" / "n" / "false" / "False" / "FALSE" / "off" / "Off" / "OFF"
    bool        = truthy / falsey

    data        = bool / text
""")



import attr


@attr.s(slots=True)
class Pos:
    index = attr.ib()
    line = attr.ib()
    column = attr.ib()


@attr.s(slots=True, repr=False)
class Source:
    filename = attr.ib()
    start = attr.ib()
    end = attr.ib()
    text = attr.ib()
    value = attr.ib()

    def __repr__(self):
        return '%sLine %s, Column %s (index %s): %r (%r)' % (
            '%s, ' % self.filename if self.filename else '',
            self.start.line, self.start.column, self.start.index, self.text, self.value,
        )



from parsimonious import NodeVisitor, Grammar, VisitationError

from . import grammars
from . import nodes
from .exceptions import OutOfContextNodeError


class TextOnlySymlParser(NodeVisitor):
    grammar = Grammar(grammars.text_only_syml_grammar)

    def reduce_children(self, children):
        children = [c for c in children if c is not None]
        if children:
            return children if len(children) > 1 else children[0]
        else:
            return None

    def visit_blank(self, node, children):
        return None

    def visit_line(self, node, children):
        indent, value, _ = children
        if value is not None:
            value.level = indent
            return value

    def generic_visit(self, node, children):
        return self.reduce_children(children)

    def get_text(self, node, children):
        return nodes.LeafNode(node, node.text)

    def visit_comment(self, node, children):
        _, text = children
        return nodes.Comment(text)

    visit_text = get_text
    visit_key = get_text

    def visit_indent(self, node, children):
        return len(node.text.replace('\t', ' ' * 4).strip('\n'))

    def visit_key_value(self, node, children):
        section, _, value = children
        section.incorporate_node(value)
        return section

    def visit_section(self, node, children):
        key, _ = children
        return nodes.KeyValue(node, key)

    def visit_list_item(self, node, children):
        _, _, value = children
        li = nodes.ListItem(node)
        li.incorporate_node(value)
        return li

    def visit_lines(self, node, children):
        root = nodes.Root(node)
        current = root

        children = self.reduce_children(children)
        if isinstance(children, nodes.LeafNode):
            children = [children]

        for child in children:
            if isinstance(child, nodes.Comment):
                current.comments.append(child)
            else:
                current = current.incorporate_node(child)
        return root

    def parse(self, *args, **kwargs):
        try:
            return super().parse(*args, **kwargs)
        except VisitationError as e:
            # Parsimonious swallows errors inside of `visit_` handlers and
            # wraps them in VisitationError cruft.
            if e.args[0].startswith('OutOfContextNodeError'):
                # Extract the original error message, ignoring the cruft.
                msg = e.args[0].split('\n\n\n')[0].split(':', 1)[1]
                raise OutOfContextNodeError(msg)
            else:
                raise  # pragma: no cover


class BooleanSymlParser(TextOnlySymlParser):
    """Syml with support for YAML-like boolean values.
    """
    grammar = Grammar(grammars.boolean_syml_grammar)

    def visit_truthy(self, node, children):
        return nodes.RawValueLeafNode(node, node.text, value=True)

    def visit_falsey(self, node, children):
        return nodes.RawValueLeafNode(node, node.text, value=False)


def parse(source_syml, filename='', raw=True, booleans=False):
    parser = BooleanSymlParser if booleans else TextOnlySymlParser
    return parser().parse(source_syml).as_data(filename, raw=raw)



from . import parser


def loads(document, **kwargs):
    return parser.parse(document, **kwargs)


def load(file_obj, **kwargs):
    return loads(file_obj.read(), **kwargs)

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
import versioneer


with open('README.rst') as README:
    setup(
        name = 'syml',
        version=versioneer.get_version(),
        author = "David Eyk",
        author_email = "david.eyk@gmail.com",
        cmdclass=versioneer.get_cmdclass(),
        description = (
            "SYML (Simple YAML-like Markup Language) is a simple markup "
            "language with similar structure to YAML, but without all the gewgaws "
            "and folderol."
        ),
        long_description = README.read(),
        url = 'https://github.com/eykd/syml',
        packages = find_packages(),
        install_requires = [
            'attrs<17',
            'parsimonious<0.8',
        ],
        keywords = ['yaml', 'markup', 'syml'],
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Developers',
            'Programming Language :: Python :: 3.5',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Topic :: Text Processing :: Markup',
        ],
    )
